class Adaptor(object):
    def __init__(self, study, encoder_weights, decoder_weights):
        self.study = study

        self.encoder_weights = encoder_weights
        self.decoder_weights = decoder_weights



